from typing import Container, Type

from django.db import models

from ..check_id import CheckId
from .ast import ModelAST
from .exceptions import MissingASTError
from .protocols import (
    ArgASTProtocol,
    FieldASTProtocol,
    ModelASTDisableCommentProtocol,
    ModelASTProtocol,
)


def get_model_ast(
    model_cls: Type[models.Model],
    meta_checks: Container[CheckId],
) -> ModelASTDisableCommentProtocol:
    return ModelAST(model_cls, meta_checks)


__all__ = [
    "get_model_ast",
    "MissingASTError",
    "ArgASTProtocol",
    "FieldASTProtocol",
    "ModelASTProtocol",
]



import pytest

try:
    import rest_framework.authtoken.serializers  # import AuthTokenSerializer into the project
    import rest_framework.serializers
except ImportError:
    pytest.skip("skipping rest_framework tests", allow_module_level=True)
from extra_checks.checks.drf_serializer_checks import (
    CheckDRFSerializerExtraKwargs,
    CheckDRFSerializerMetaAttribute,
    _get_serializers_to_check,
    check_drf_serializers,
)
from tests.example.serializers import (
    ArticleSerializer,
    AuthorSerializer,
    DisableCheckSerializer,
)


@pytest.fixture
def test_case(test_case):
    return test_case.handler(check_drf_serializers)


def test_model_serializer_extra_kwargs(test_case):
    messages = (
        test_case.settings({"checks": [CheckDRFSerializerExtraKwargs.Id.value]})
        .check(CheckDRFSerializerExtraKwargs)
        .serializers(ArticleSerializer)
        .run()
    )
    assert not messages
    messages = test_case.serializers(AuthorSerializer).run()
    assert len(messages) == 1
    assert messages[0].id == CheckDRFSerializerExtraKwargs.Id.name


def test_model_serializer_meta_attribute(test_case):
    messages = (
        test_case.settings(
            {
                "checks": [
                    {
                        "id": CheckDRFSerializerMetaAttribute.Id.value,
                        "attrs": ["read_only_fields"],
                    }
                ]
            }
        )
        .check(CheckDRFSerializerMetaAttribute)
        .serializers(ArticleSerializer)
        .run()
    )
    assert not messages
    messages = test_case.serializers(AuthorSerializer).run()
    assert len(messages) == 1
    assert messages[0].id == CheckDRFSerializerMetaAttribute.Id.name


def test_get_serializers_to_check():
    ss, ms = _get_serializers_to_check()
    module = "tests.example.serializers"
    assert all(s.__module__ == module for s in ss)
    assert all(m.__module__ == module for m in ms)


def test_get_serializers_to_check_include_apps(settings):
    settings.INSTALLED_APPS += ["rest_framework.authtoken", "rest_framework"]
    ss, ms = (list(s) for s in _get_serializers_to_check(["rest_framework.authtoken"]))
    assert not ms
    assert len(ss) == 1
    assert ss[0] is rest_framework.authtoken.serializers.AuthTokenSerializer

    ss, ms = (list(s) for s in _get_serializers_to_check(["rest_framework"]))
    assert len(ms) == 1
    assert ms[0] is rest_framework.serializers.HyperlinkedModelSerializer
    # FIXME: the lines below actually a bug
    # include_apps must filter out rest_framework.authtoken app
    # but it's nested into rest_framework app that present in the
    # include_apps so it's included too
    assert len(ss) == 1
    assert ss[0] is rest_framework.authtoken.serializers.AuthTokenSerializer


def test_drf_ignore_meta_checks(test_case):
    messages = (
        test_case.settings({"checks": [CheckDRFSerializerExtraKwargs.Id.value]})
        .check(CheckDRFSerializerExtraKwargs)
        .serializers(DisableCheckSerializer)
        .run()
    )
    assert not messages



import sys
from typing import Any, Iterable, Optional, Tuple

if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol

from django.db import models


class ArgASTProtocol(Protocol):
    @property
    def is_callable(self) -> bool:
        ...

    @property
    def callable_func_name(self) -> Optional[str]:
        ...

    def get_call_first_args(self) -> Any:
        ...


class FieldASTProtocol(Protocol):
    def get_arg(self, name: str) -> Optional[ArgASTProtocol]:
        ...


class ModelASTProtocol(Protocol):
    @property
    def field_nodes(
        self,
    ) -> Iterable[Tuple[models.fields.Field, "FieldASTDisableCommentProtocol"]]:
        ...

    def has_meta_var(self, name: str) -> bool:
        ...


class DisableCommentProtocol(Protocol):
    def is_disabled_by_comment(self, check_id: str) -> bool:
        ...


class ModelASTDisableCommentProtocol(
    ModelASTProtocol, DisableCommentProtocol, Protocol
):
    ...


class FieldASTDisableCommentProtocol(
    FieldASTProtocol, DisableCommentProtocol, Protocol
):
    ...



import ast
from functools import partial
from typing import (
    TYPE_CHECKING,
    Callable,
    Container,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)

from django.db import models
from django.utils.functional import SimpleLazyObject

from extra_checks.check_id import CheckId

from .exceptions import MissingASTError
from .protocols import (
    ArgASTProtocol,
    DisableCommentProtocol,
    FieldASTProtocol,
    ModelASTProtocol,
)
from .source_provider import SourceProvider

if TYPE_CHECKING:
    cached_property = property
else:
    from django.utils.functional import cached_property


class ModelAST(DisableCommentProtocol, ModelASTProtocol):
    def __init__(self, model_cls: Type[models.Model], meta_checks: Container[CheckId]):
        self.model_cls = model_cls
        self.meta_checks = meta_checks
        self._assignment_nodes: List[ast.Assign] = []
        self._meta: Optional[ast.ClassDef] = None

    @cached_property
    def _source_provider(self) -> SourceProvider:
        return SourceProvider(self.model_cls)

    @cached_property
    def _nodes(self) -> Iterator[ast.AST]:
        if self._source_provider.source is None:
            return iter([])
        return iter(ast.parse(self._source_provider.source).body[0].body)  # type: ignore

    def _parse(self, predicate: Optional[Callable[[ast.AST], bool]] = None) -> None:
        try:
            for node in self._nodes:
                if predicate and predicate(node):
                    self._meta = cast(ast.ClassDef, node)
                    break
                if isinstance(node, ast.Assign):
                    self._assignment_nodes.append(node)
        except StopIteration:
            return
        return

    @cached_property
    def _meta_node(self) -> Optional[ast.ClassDef]:
        if not self._meta:
            self._parse(
                lambda node: isinstance(node, ast.ClassDef) and node.name == "Meta"
            )
        return self._meta

    @cached_property
    def _meta_vars(self) -> Dict[str, ast.Assign]:
        data: Dict[str, ast.Assign] = {}
        if not self._meta_node:
            return data
        for node in ast.iter_child_nodes(self._meta_node):
            if isinstance(node, ast.Assign) and isinstance(node.targets[0], ast.Name):
                data[node.targets[0].id] = node
        return data

    @cached_property
    def _assignments(self) -> Dict[str, ast.Assign]:
        self._parse()
        result = {}
        for node in self._assignment_nodes:
            if isinstance(node.targets[0], ast.Name):
                result[node.targets[0].id] = node
        return result

    @cached_property
    def field_nodes(self) -> Iterable[Tuple[models.fields.Field, "FieldAST"]]:
        for field in self.model_cls._meta.get_fields(include_parents=False):
            if isinstance(field, models.Field):
                yield field, cast(
                    FieldAST, SimpleLazyObject(partial(get_field_ast, self, field))
                )

    def has_meta_var(self, name: str) -> bool:
        return name in self._meta_vars

    def is_disabled_by_comment(self, check_id: str) -> bool:
        check = CheckId.find_check(check_id)
        if check in self.meta_checks:
            if not self._meta_node:
                # class Meta is not defined on model
                return False
            return check in self._source_provider.get_disabled_checks_for_line(
                self._meta_node.lineno
            )
        return check in self._source_provider.get_disabled_checks_for_line(1)


def get_field_ast(model_ast: ModelAST, field: models.Field) -> "FieldAST":
    try:
        return FieldAST(
            model_ast._assignments[field.name], field, model_ast._source_provider
        )
    except KeyError:
        raise MissingASTError()


class ArgAST(ArgASTProtocol):
    def __init__(self, node: ast.AST):
        self._node = node

    @cached_property
    def is_callable(self) -> bool:
        return isinstance(self._node, ast.Call)

    @cached_property
    def callable_func_name(self) -> Optional[str]:
        return (
            getattr(self._node.func, "id", None)
            if isinstance(self._node, ast.Call)
            else None
        )

    def get_call_first_args(self) -> str:
        return self._node.args[0].s  # type: ignore


class FieldAST(DisableCommentProtocol, FieldASTProtocol):
    def __init__(
        self, node: ast.Assign, field: models.Field, source_provider: SourceProvider
    ):
        self._node = node
        self._field = field
        self._source_provider = source_provider

    @cached_property
    def _args(self) -> List[ast.expr]:
        return self._node.value.args  # type: ignore

    @cached_property
    def _kwargs(self) -> Dict[str, ast.keyword]:
        return {kw.arg: kw for kw in self._node.value.keywords if kw.arg}  # type: ignore

    def get_arg(self, name: str) -> Optional[ArgASTProtocol]:
        if name == "verbose_name":
            return ArgAST(self._verbose_name) if self._verbose_name else None
        return ArgAST(self._kwargs[name].value) if name in self._kwargs else None

    @cached_property
    def _verbose_name(self) -> Union[None, ast.Constant, ast.Call]:
        result = getattr(self._kwargs.get("verbose_name"), "value", None)
        if result:
            return result
        if isinstance(self._field, models.fields.related.RelatedField):
            return None
        if self._args:
            node = self._args[0]
            if isinstance(node, ast.Call) and hasattr(node.func, "id"):
                return node
            elif isinstance(node, (ast.Constant, ast.Str)):
                return node
        return None

    def is_disabled_by_comment(self, check_id: str) -> bool:
        return CheckId.find_check(
            check_id
        ) in self._source_provider.get_disabled_checks_for_line(self._node.lineno)



import site
from abc import abstractmethod
from typing import TYPE_CHECKING, Any, Iterable, Iterator, List, Optional, Type, Union

import django.core.checks
from django import forms
from django.apps import apps
from django.contrib.admin.sites import all_sites
from django.db import models
from django.db.models.options import DEFAULT_NAMES as META_ATTRS

from .. import CheckId
from ..ast import ModelASTProtocol, get_model_ast
from ..forms import AttrsForm, BaseCheckForm
from ..registry import ChecksConfig, registry
from .base_checks import BaseCheck

if TYPE_CHECKING:
    from .model_field_checks import CheckModelField


def _get_models_to_check(
    *,
    app_configs: Optional[List[Any]] = None,
    include_apps: Optional[Iterable[str]] = None,
) -> Iterator[Type[models.Model]]:
    apps = django.apps.apps.get_app_configs() if app_configs is None else app_configs
    if include_apps is not None:
        for app in apps:
            if app.name in include_apps:
                yield from app.get_models()
        return
    for app in apps:
        if not any(app.path.startswith(path) for path in set(site.PREFIXES)):
            yield from app.get_models()


@registry.add_handler(django.core.checks.Tags.models)
def check_models(
    checks: Iterable[Union["CheckModel", "CheckModelField", "CheckModelMeta"]],
    config: ChecksConfig,
    app_configs: Optional[List[Any]] = None,
    **kwargs: Any,
) -> Iterator[Any]:
    model_checks: List[Union[CheckModel, CheckModelMeta]] = []
    field_checks = []
    meta_checks = []
    for check in checks:
        if isinstance(check, CheckModelMeta):
            meta_checks.append(check)
            model_checks.append(check)
        elif isinstance(check, CheckModel):
            model_checks.append(check)
        else:
            field_checks.append(check)
    if not model_checks and not field_checks:
        return
    for model in _get_models_to_check(
        app_configs=app_configs, include_apps=config.include_apps
    ):
        model_ast = get_model_ast(model, [c.Id for c in meta_checks])
        for check in model_checks:
            yield from check(model, ast=model_ast)
        if field_checks:
            for field, field_ast in model_ast.field_nodes:
                for check in field_checks:
                    yield from check(field, ast=field_ast, model=model)


class CheckModel(BaseCheck):
    @abstractmethod
    def apply(
        self, model: Type[models.Model], ast: ModelASTProtocol
    ) -> Iterator[django.core.checks.CheckMessage]:
        raise NotImplementedError()


class CheckModelMeta(BaseCheck):
    @abstractmethod
    def apply(
        self, model: Type[models.Model], ast: ModelASTProtocol
    ) -> Iterator[django.core.checks.CheckMessage]:
        raise NotImplementedError()


@registry.register(django.core.checks.Tags.models)
class CheckModelAttribute(CheckModel):
    Id = CheckId.X010
    settings_form_class = AttrsForm

    def __init__(self, attrs: List[str], **kwargs: Any) -> None:
        self.attrs = attrs
        super().__init__(**kwargs)

    def apply(
        self, model: Type[models.Model], ast: ModelASTProtocol
    ) -> Iterator[django.core.checks.CheckMessage]:
        for attr in self.attrs:
            if (
                not model._meta.abstract
                and not model._meta.proxy
                and not hasattr(model, attr)
            ):
                yield self.message(
                    f'Each model must specify "{attr}" attribute.',
                    hint=f'Set "{attr}" attribute.',
                    obj=model,
                )


@registry.register(django.core.checks.Tags.models)
class CheckModelMetaAttribute(CheckModelMeta):
    Id = CheckId.X011

    class MetaAttrsForm(BaseCheckForm):
        attrs = forms.MultipleChoiceField(choices=[(o, o) for o in META_ATTRS])

    settings_form_class = MetaAttrsForm

    def __init__(self, attrs: List[str], **kwargs: Any) -> None:
        self.attrs = attrs
        super().__init__(**kwargs)

    def apply(
        self, model: Type[models.Model], ast: ModelASTProtocol
    ) -> Iterator[django.core.checks.CheckMessage]:
        for attr in self.attrs:
            if (
                not model._meta.abstract
                and not model._meta.proxy
                and not ast.has_meta_var(attr)
            ):
                yield self.message(
                    f'Each model must specify "{attr}" attribute in its Meta.',
                    hint=f'Set "{attr}" attribute in Meta.',
                    obj=model,
                )


@registry.register(django.core.checks.Tags.models)
class CheckModelAdmin(CheckModel):
    Id = CheckId.X012

    class AdminForm(BaseCheckForm):
        def clean(self) -> dict:
            if not apps.is_installed("django.contrib.admin"):
                raise forms.ValidationError(
                    "django.contrib.admin must be in INSTALLED_APPS."
                )
            return super().clean()

    settings_form_class = AdminForm

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.models_with_admin = set()
        for admin_site in all_sites:
            for model_cls, admin_cls in admin_site._registry.items():
                self.models_with_admin.add(model_cls)
                for inline in admin_cls.inlines:
                    self.models_with_admin.add(inline.model)

    def apply(
        self, model: Type[models.Model], ast: ModelASTProtocol
    ) -> Iterator[django.core.checks.CheckMessage]:
        if model not in self.models_with_admin:
            yield self.message("The model is not registered in admin.", obj=model)


@registry.register(django.core.checks.Tags.models)
class CheckNoUniqueTogether(CheckModelMeta):
    Id = CheckId.X013

    def apply(
        self, model: Type[models.Model], ast: ModelASTProtocol
    ) -> Iterator[django.core.checks.CheckMessage]:
        if ast.has_meta_var("unique_together"):
            yield self.message(
                "Use UniqueConstraint with the constraints option instead.",
                obj=model,
            )


@registry.register(django.core.checks.Tags.models)
class CheckNoIndexTogether(CheckModelMeta):
    Id = CheckId.X014

    def apply(
        self, model: Type[models.Model], ast: ModelASTProtocol
    ) -> Iterator[django.core.checks.CheckMessage]:
        if ast.has_meta_var("index_together"):
            yield self.message("Use the indexes option instead", obj=model)



from typing import Iterable, Iterator, Optional, Set, Type, TypeVar

TBase = TypeVar("TBase")


def collect_subclasses(
    bases: Iterable[Type[TBase]],
    visited: Optional[Set[Type[TBase]]] = None,
) -> Iterator[Type[TBase]]:
    visited = visited or set()
    for cls in bases:
        if cls not in visited:
            visited.add(cls)
            yield from collect_subclasses(cls.__subclasses__(), visited)
            yield cls



from django.core import checks

from extra_checks import CheckId
from extra_checks.checks import (
    CheckFieldFileUploadTo,
    CheckModelAttribute,
    CheckModelMetaAttribute,
)
from extra_checks.forms import ConfigForm
from extra_checks.registry import ChecksConfig
from tests.example.models import Article, Author


def test_config_form():
    form = ConfigForm(
        data={
            "checks": [
                {
                    "id": CheckModelAttribute.Id.value,
                    "level": "WARNING",
                    "attrs": ["db_table"],
                }
            ]
        }
    )
    assert form.is_valid(
        {CheckModelAttribute.Id: CheckModelAttribute.settings_form_class}
    )
    assert form.cleaned_data == {
        "checks": {
            CheckModelAttribute.Id.value: {
                "level": checks.WARNING,
                "attrs": ["db_table"],
            }
        },
    }

    form = ConfigForm(data={"checks": [CheckFieldFileUploadTo.Id.value]})
    assert form.is_valid(
        {CheckFieldFileUploadTo.Id: CheckFieldFileUploadTo.settings_form_class}
    )
    assert form.cleaned_data == {
        "checks": {CheckFieldFileUploadTo.Id.value: {"level": None}},
    }

    form = ConfigForm(data={})
    assert form.is_valid({})
    assert form.cleaned_data == {"checks": {}}

    form = ConfigForm(
        data={"checks": [{"id": CheckModelAttribute.Id.value, "attrs": ["db_table"]}]}
    )
    assert form.is_valid(
        {CheckModelAttribute.Id: CheckModelAttribute.settings_form_class}
    )
    assert form.cleaned_data == {
        "checks": {
            CheckModelAttribute.Id.value: {"level": None, "attrs": ["db_table"]}
        },
    }

    form = ConfigForm(data={"checks": [{5}]})
    assert not form.is_valid(
        {CheckModelAttribute.Id: CheckModelAttribute.settings_form_class}
    )
    assert form.errors == {
        "checks": ["{5} is not one of the available types."],
    }

    form = ConfigForm(data={"checks": CheckModelAttribute.Id.value})
    assert not form.is_valid(
        {CheckModelAttribute.Id: CheckModelAttribute.settings_form_class}
    )
    assert form.errors == {
        "checks": ["Enter a list of values."],
    }

    form = ConfigForm(data={"checks": ["X000"]})
    assert not form.is_valid(
        {CheckFieldFileUploadTo.Id: CheckFieldFileUploadTo.settings_form_class}
    )
    assert form.errors == {"checks": ["X000 is not one of the available checks."]}

    form = ConfigForm(
        data={
            "checks": [
                {
                    "id": CheckModelMetaAttribute.Id.value,
                    "level": "FAKE",
                    "attrs": ["random"],
                }
            ]
        }
    )
    assert not form.is_valid(
        {CheckModelMetaAttribute.Id: CheckModelMetaAttribute.settings_form_class}
    )
    assert form.errors == {
        "checks": {
            CheckModelMetaAttribute.Id.value: {
                "attrs": [
                    "Select a valid choice. random is not one of the available choices."
                ],
                "level": [
                    "Select a valid choice. FAKE is not one of the available choices."
                ],
            },
        }
    }


def test_config_include_apps():
    form = ConfigForm(data={"checks": []})
    assert form.is_valid({})
    assert form.cleaned_data == {"checks": {}}

    form = ConfigForm(data={"include_apps": [], "checks": []})
    assert form.is_valid({})
    assert form.cleaned_data == {"include_apps": [], "checks": {}}

    form = ConfigForm(data={"include_apps": ["tests.example"], "checks": []})
    assert form.is_valid({})
    assert form.cleaned_data == {"include_apps": ["tests.example"], "checks": {}}


def test_config_build_ignored():
    ignored, errors = ChecksConfig._build_ignored(
        {
            Article: {"X010", "random-test-name"},
            Author: {"model-meta-attribute", CheckId.X050},
        }
    )
    assert errors == [
        "Unknown check (random-test-name) provided to the 'ignore_checks'."
    ]
    assert ignored == {
        CheckId.X010: {Article},
        CheckId.X011: {Author},
        CheckId.X050: {Author},
    }



from extra_checks.utils import collect_subclasses


def test_collect_subclasses():
    """test that we collect all descendants"""

    class Base:
        pass

    class One(Base):
        pass

    class Two(Base):
        pass

    class Three(One, Two):
        pass

    serializers = collect_subclasses(Base.__subclasses__())
    assert set(serializers) == {One, Two, Three}



from django.contrib import admin

from .models import Article, Author


class ArticleInline(admin.TabularInline):
    model = Article


@admin.register(Author)
class AuthorAdmin(admin.ModelAdmin):
    inlines = [
        ArticleInline,
    ]



import django.db.models
import pytest

from extra_checks.checks import model_checks, model_field_checks
from tests.example import models


@pytest.fixture
def test_case(test_case):
    return test_case.handler(model_checks.check_models)


def test_ignore_model_check(test_case):
    messages = (
        test_case.models(models.DisableCheckModel)
        .settings(
            {
                "checks": [
                    {
                        "id": model_checks.CheckModelAttribute.Id.value,
                        "attrs": ["site"],
                    },
                    model_field_checks.CheckFieldTextNull.Id.value,
                    model_checks.CheckNoUniqueTogether.Id.value,
                ]
            }
        )
        .check(
            model_checks.CheckModelAttribute,
            model_field_checks.CheckFieldTextNull,
            model_checks.CheckNoUniqueTogether,
        )
        .run()
    )
    assert not messages


def test_ignore_many_model_check(test_case):
    messages = (
        test_case.models(models.DisableManyChecksModel)
        .settings(
            {
                "checks": [
                    model_field_checks.CheckFieldTextNull.Id.value,
                    model_field_checks.CheckFieldVerboseName.Id.value,
                    model_checks.CheckNoUniqueTogether.Id.value,
                    model_checks.CheckNoIndexTogether.Id.value,
                ]
            }
        )
        .check(
            model_field_checks.CheckFieldTextNull,
            model_field_checks.CheckFieldVerboseName,
            model_checks.CheckNoUniqueTogether,
            model_checks.CheckNoIndexTogether,
        )
        .run()
    )
    assert not messages


def test_field_skipif(test_case):
    def skipif(field, *args, **kwargs):
        return isinstance(field, django.db.models.ImageField)

    messages = (
        test_case.settings(
            {
                "checks": [
                    {
                        "id": model_field_checks.CheckFieldFileUploadTo.Id.value,
                        "skipif": skipif,
                    }
                ],
            }
        )
        .models(models.ModelFieldFileUploadTo)
        .check(model_field_checks.CheckFieldFileUploadTo)
        .run()
    )
    assert len(messages) == 1
    assert {m.obj.name for m in messages} == {"file_fail"}


def test_model_skipif(test_case):
    def skipif(model, *args, **kwargs):
        return not any(
            isinstance(f, django.db.models.DateTimeField)
            for f in model._meta.get_fields()
        )

    messages = (
        test_case.settings(
            {
                "checks": [
                    {
                        "id": model_checks.CheckModelMetaAttribute.Id.value,
                        "attrs": ["get_latest_by"],
                        "skipif": skipif,
                    },
                ],
            }
        )
        .models(models.Article, models.Author)
        .check(model_checks.CheckModelMetaAttribute)
        .run()
    )
    assert len(messages) == 0



import warnings
from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Iterator,
    List,
    Optional,
    Set,
    Type,
)

import django.core.checks

from .. import CheckId, forms
from ..ast.protocols import DisableCommentProtocol

MESSAGE_MAP = {
    django.core.checks.DEBUG: django.core.checks.Debug,
    django.core.checks.INFO: django.core.checks.Info,
    django.core.checks.WARNING: django.core.checks.Warning,
    django.core.checks.ERROR: django.core.checks.Error,
    django.core.checks.CRITICAL: django.core.checks.Critical,
}


class BaseCheck(ABC):
    Id: CheckId
    settings_form_class: ClassVar[Type[forms.BaseCheckForm]] = forms.BaseCheckForm
    level = django.core.checks.WARNING
    deprecation_warnings: List[str] = []

    def __init__(
        self,
        level: Optional[int] = None,
        ignore_objects: Optional[Set[Any]] = None,
        ignore_types: Optional[set] = None,
        skipif: Optional[Callable] = None,
    ) -> None:
        self.level = level or self.level
        self.ignore_objects = ignore_objects or set()
        self.ignore_types = ignore_types or set()
        self.skipif = skipif
        for warning in self.deprecation_warnings:
            warnings.warn(warning, FutureWarning)

    def __call__(
        self, obj: Any, ast: Optional[DisableCommentProtocol] = None, **kwargs: Any
    ) -> Iterator[django.core.checks.CheckMessage]:
        if not self.is_ignored(obj):
            for error in self.apply(obj, ast=ast, **kwargs):
                if not ast or (error.id and not ast.is_disabled_by_comment(error.id)):
                    yield error

    def is_ignored(self, obj: Any) -> bool:
        if self.skipif and self.skipif(obj):
            return True
        return obj in self.ignore_objects or type(obj) in self.ignore_types

    def message(
        self, message: str, hint: Optional[str] = None, obj: Any = None
    ) -> django.core.checks.CheckMessage:
        return MESSAGE_MAP[self.level](
            message + f" [{self.Id.value}]", hint=hint, obj=obj, id=self.Id.name
        )

    @abstractmethod
    def apply(
        self, *args: Any, **kwargs: Any
    ) -> Iterator[django.core.checks.CheckMessage]:
        raise NotImplementedError()


if TYPE_CHECKING:
    BaseCheckMixin = BaseCheck
else:
    BaseCheckMixin = object



from rest_framework import serializers

from .models import Article, Author


class ArticleSerializer(serializers.ModelSerializer):
    class Meta:
        model = Article
        read_only_fields: list = []


class AuthorSerializer(serializers.ModelSerializer):
    first_name = serializers.CharField()

    class Meta:
        model = Author
        extra_kwargs = {"first_name": {"read_only": True}}


class DisableCheckSerializer(serializers.ModelSerializer):
    first_name = serializers.CharField()

    def method_with_meta_class(self):
        class Meta:
            ...

    # extra-checks-disable-next-line drf-model-serializer-extra-kwargs
    class Meta:
        model = Author
        extra_kwargs = {"first_name": {"read_only": True}}



from django.apps import AppConfig

from . import checks  # noqa
from .registry import registry


class ExtraChecksConfig(AppConfig):
    name = "extra_checks"

    def ready(self) -> None:
        super().ready()
        registry.bind()



import pytest

from extra_checks.checks.self_checks import CheckConfig
from extra_checks.registry import Registry


@pytest.fixture
def registry() -> Registry:
    registry = Registry()
    registry._register(["extra_checks_selfcheck"], CheckConfig)
    return registry


class TestCase:
    TEST_TAG = "__test_case_check__"

    def __init__(self, settings, monkeypatch, registry):
        self._settings = settings
        self._monkeypatch = monkeypatch
        self._registry = registry

    def settings(self, settings):
        self._settings.EXTRA_CHECKS = settings
        return self

    def check(self, *checks):
        for check in checks:
            self._registry._register([self.TEST_TAG], check)
        return self

    def handler(self, handler):
        self._registry._add_handler(self.TEST_TAG, handler)
        return self

    def run(self):
        self._registry.enabled_checks = {}
        handlers = self._registry.bind()
        assert (
            self._registry.is_healthy
        ), f"Settings has errors: {self._registry._config.errors.as_text()}"
        return list(handlers[self.TEST_TAG]())

    def models(self, *models):
        self._monkeypatch.setattr(
            "extra_checks.checks.model_checks._get_models_to_check",
            lambda *a, **kw: (m for m in models),
        )
        return self

    def serializers(self, *serializers):
        import rest_framework.serializers

        n, m = [], []
        for s in serializers:
            if issubclass(s, rest_framework.serializers.ModelSerializer):
                m.append(s)
            else:
                n.append(s)
        self._monkeypatch.setattr(
            "extra_checks.checks.drf_serializer_checks._get_serializers_to_check",
            lambda *a, **kw: (n, m),
        )
        return self


@pytest.fixture
def test_case(settings, monkeypatch, registry):
    return TestCase(settings, monkeypatch, registry)



from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.contrib.sites.models import Site
from django.db import models
from django.utils.translation import gettext_lazy

_ = gettext_lazy


if True:
    # test that indentation doesn't break ast parser
    class Author(models.Model):
        first_name = models.CharField(max_length=100)
        last_name = models.CharField(max_length=100)


class Article(models.Model):
    site = models.ForeignKey(Site, verbose_name="site", on_delete=models.CASCADE)
    title = models.CharField("article title", max_length=100)
    text = models.TextField(verbose_name="article text")
    author = models.ForeignKey(
        Author, related_name="articles", on_delete=models.CASCADE
    )
    created = models.DateTimeField(auto_now=True)

    class Meta:
        verbose_name = "Site Article"
        get_latest_by = ["created"]


class NestedField(models.Field):
    """Resembles fields like postgres ArrayField."""

    def __init__(self, base_field, **kwargs):
        self.base_field = base_field
        super().__init__(**kwargs)


class ModelFieldVerboseName(models.Model):
    first_arg_name = models.CharField("first arg name [test]", max_length=32)
    kwarg_name = models.CharField(verbose_name="kwarg name [test]", max_length=32)
    arg_gettext = models.CharField(_("arg name [test]"), max_length=32)
    kwargs_gettext = models.CharField(
        verbose_name=_("kwarg name [test]"), max_length=32
    )
    gettext_case = models.CharField(verbose_name=_("Kwarg Name [test]"), max_length=32)
    gettext = models.CharField(
        verbose_name=gettext_lazy("kwarg name [test]"), max_length=32
    )
    name_related = models.ForeignKey(
        Article,
        on_delete=models.CASCADE,
        related_name="+",
        verbose_name="name related test [X050]",
    )
    nested_field = NestedField(
        models.CharField(max_length=100), verbose_name="nested field [X050]"
    )
    no_name = models.CharField(max_length=32)
    no_name_related = models.ForeignKey(
        Article, on_delete=models.CASCADE, related_name="+"
    )
    no_name_nested_field = NestedField(models.CharField(max_length=100))


class ModelFieldHelpTextGettext(models.Model):
    text = models.CharField(max_length=10, help_text=_("my help text"))
    text_fail = models.CharField(max_length=10, help_text="my help text")


class ModelFieldFileUploadTo(models.Model):
    image = models.ImageField(upload_to="path/to/media")
    file = models.FileField(upload_to="path/to/files")
    image_fail = models.ImageField()
    file_fail = models.FileField()


class CustomTextField(models.TextField):
    pass


class ModelFieldTextNull(models.Model):
    text = models.TextField()
    chars = models.CharField(max_length=32)
    custom = CustomTextField()
    text_fail = models.TextField(null=True)
    chars_fail = models.CharField(null=True, max_length=32)
    custom_fail = CustomTextField(null=True)


class ModelFieldNullFalse(models.Model):
    myfield = models.IntegerField()
    myfield_fail = models.IntegerField(null=False)
    null_fail = models.NullBooleanField()


class ModelFieldNullDefault(models.Model):
    myfield = models.IntegerField(default=None)
    myfield_fail = models.IntegerField(null=True, default=None)


class ModelFieldForeignKeyIndex(models.Model):
    article = models.ForeignKey(
        Article, on_delete=models.CASCADE, related_name="+", db_index=True
    )
    author = models.ForeignKey(Author, on_delete=models.CASCADE, related_name="+")
    another_article = models.ForeignKey(
        Article, on_delete=models.CASCADE, related_name="+"
    )
    another_author = models.ForeignKey(
        Article, on_delete=models.CASCADE, related_name="+", db_index=True
    )
    field_one = models.ForeignKey(
        ModelFieldTextNull, on_delete=models.CASCADE, related_name="+"
    )
    field_two = models.ForeignKey(
        ModelFieldNullFalse, on_delete=models.CASCADE, related_name="+", db_index=True
    )
    field_three = models.ForeignKey(
        ModelFieldNullFalse, on_delete=models.CASCADE, related_name="+"
    )
    field_in_indexes = models.ForeignKey(
        ModelFieldNullFalse, on_delete=models.CASCADE, related_name="+"
    )
    field_index_desc = models.ForeignKey(
        ModelFieldNullFalse, on_delete=models.CASCADE, related_name="+"
    )

    class Meta:
        unique_together = [("author", "article")]
        index_together = ("field_one", "field_two")
        constraints = [
            models.UniqueConstraint(
                fields=("author", "field_three"), name="fi_author_field_unique"
            )
        ]
        indexes = [
            models.Index(fields=("field_in_indexes",)),
            models.Index(fields=("field_in_indexes", "-field_index_desc")),
        ]


class GenericKeyOne(models.Model):
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    service = GenericForeignKey("content_type", "object_id")


class GenericKeyTwo(models.Model):
    ones = GenericRelation("GenericKeyOne")


class ChoicesConstraint(models.Model):
    non_choice = models.IntegerField()
    empty = models.IntegerField(choices=[])
    partial = models.CharField(
        choices=[("S", "simple"), ("C", "complex")], max_length=1
    )
    missed = models.IntegerField(choices=[(1, "One"), (2, "Two")])
    covered = models.CharField(choices=[("A", "a"), ("B", "b")])
    blank = models.CharField(choices=[("A", "a"), ("B", "b")], blank=True)
    blank_missed = models.CharField(choices=[("A", "a"), ("B", "b")], blank=True)
    blank_included = models.CharField(
        choices=[("A", "a"), ("B", "b"), ("", "---")], blank=True
    )
    grouped = models.IntegerField(
        choices=[("g1", ((1, "1"), (2, "2"))), ("g2", ((3, "3"),))]
    )
    integer_blank = models.IntegerField(
        choices=[(1, "One"), (2, "Two")], blank=True, null=True
    )
    integer_blank_invalid = models.IntegerField(
        choices=[(1, "One"), (2, "Two")], blank=True, null=True
    )

    class Meta:
        constraints = [
            models.CheckConstraint(
                name="partial_valid", check=models.Q(name__in=["S"])
            ),
            models.CheckConstraint(
                name="covered_valid", check=models.Q(covered__in=("A", "B"))
            ),
            models.CheckConstraint(
                name="blank_valid", check=models.Q(blank__in=("A", "B", ""))
            ),
            models.CheckConstraint(
                name="blank_missed_valid", check=models.Q(blank_missed__in=("A", "B"))
            ),
            models.CheckConstraint(
                name="grouped_valid", check=models.Q(grouped__in=(1, 2, 3))
            ),
            models.CheckConstraint(
                name="integer_blank", check=models.Q(integer_blank__in=(1, 2))
            ),
            models.CheckConstraint(
                name="integer_blank_invalid",
                check=models.Q(integer_blank_invalid__in=(1, 2, "")),
            ),
        ]


# model checks can be disabled by comment right before the model class
# if your model is decorated than comment must be placed between
# decorator and class definition. eg:
# >>> @mydecorator
# >>> # extra-checks-disable-next-line model-attribute
# >>> class MyModel:...
#
# extra-checks-disable-next-line model-attribute
class DisableCheckModel(models.Model):
    not_site = models.ForeignKey(Site, verbose_name="site", on_delete=models.CASCADE)
    # extra-checks-disable-next-line field-text-null
    text_fail = models.TextField(null=True)

    # extra-checks-disable-next-line no-unique-together
    class Meta:
        unique_together = ("text_fail", "no_site")


class DisableManyChecksModel(models.Model):
    # disable two checks
    # extra-checks-disable-next-line field-text-null, field-verbose-name
    text_fail = models.TextField(null=True)
    # this disables all checks
    # extra-checks-disable-next-line
    text_fail2 = models.TextField(null=True)

    # disable two checks with separate comments
    # extra-checks-disable-next-line X014
    # extra-checks-disable-next-line no-unique-together
    class Meta:
        unique_together = ("text_fail", "text_fail2")
        index_together = ("text_fail", "text_fail2")



import importlib
import site
from abc import abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)

import django.core.checks
from rest_framework.serializers import ModelSerializer, Serializer

from ..ast.protocols import DisableCommentProtocol
from ..ast.source_provider import SourceProvider
from ..check_id import CheckId
from ..forms import AttrsForm
from ..registry import ChecksConfig, registry
from ..utils import collect_subclasses
from .base_checks import BaseCheck

if TYPE_CHECKING:
    cached_property = property
else:
    from django.utils.functional import cached_property


class DisableCommentProvider(DisableCommentProtocol):
    def __init__(self, serializer_class: Type[Serializer]):
        self.serializer_class = serializer_class

    @cached_property
    def _source_provider(self) -> SourceProvider:
        return SourceProvider(self.serializer_class)

    def _get_line(self) -> Optional[int]:
        return 1

    def is_disabled_by_comment(self, check_id: str) -> bool:
        check = CheckId.find_check(check_id)
        line = self._get_line()
        return (
            line is not None
            and check in self._source_provider.get_disabled_checks_for_line(line)
        )


class DisableMetaCommentProvider(DisableCommentProvider):
    def _get_line(self) -> Optional[int]:
        lines = (self._source_provider.source or "").splitlines()
        # find line starting with `class Meta` and lowest indent
        try:
            return sorted(
                (
                    (i, line)
                    for i, line in enumerate(lines, 1)
                    if line.strip().startswith(("class Meta(", "class Meta:"))
                ),
                key=lambda a: a[1].find("class Meta"),
            )[0][0]
        except StopIteration:
            return None


def _filter_app_serializers(
    serializers: Iterable[Type[Serializer]],
    include_apps: Optional[Iterable[str]] = None,
) -> Iterator[Type[Serializer]]:
    site_prefixes = set(site.PREFIXES)
    if include_apps is not None:
        app_paths = {
            a.path for a in django.apps.apps.get_app_configs() if a.name in include_apps
        }
        for s in serializers:
            module = importlib.import_module(s.__module__)
            if any(
                module.__file__ and module.__file__.startswith(path)
                for path in app_paths
            ):
                yield s
        return
    for s in serializers:
        module = importlib.import_module(s.__module__)
        if not any(
            module.__file__ and module.__file__.startswith(path)
            for path in site_prefixes
        ):
            yield s


def _get_serializers_to_check(
    include_apps: Optional[Iterable[str]] = None,
) -> Tuple[Iterator[Type[Serializer]], Iterator[Type[ModelSerializer]]]:
    serializer_classes = _filter_app_serializers(
        collect_subclasses(
            s for s in Serializer.__subclasses__() if s is not ModelSerializer  # type: ignore
        ),
        include_apps,
    )
    model_serializer_classes = _filter_app_serializers(
        collect_subclasses(ModelSerializer.__subclasses__()), include_apps
    )
    return (
        serializer_classes,
        cast(Iterator[Type[ModelSerializer]], model_serializer_classes),
    )


@registry.add_handler("extra_checks_drf_serializer")
def check_drf_serializers(
    checks: Iterable[
        Union[
            "CheckDRFSerializer",
            "CheckDRFModelSerializer",
            "CheckDRFModelSerializerMeta",
        ]
    ],
    config: ChecksConfig,
    app_configs: Optional[List[Any]] = None,
    **kwargs: Any,
) -> Iterator[Any]:
    model_serializer_checks = []
    model_meta_serializer_checks = []
    serializer_checks = []
    for check in checks:
        if isinstance(check, CheckDRFModelSerializerMeta):
            model_meta_serializer_checks.append(check)
        elif isinstance(check, CheckDRFModelSerializer):
            model_serializer_checks.append(check)
        else:
            serializer_checks.append(check)
    s_classes, m_classes = _get_serializers_to_check(config.include_apps)
    for s in s_classes:
        comment_provider = DisableCommentProvider(s)
        for check in serializer_checks:
            yield from check(s, comment_provider)
    for s in m_classes:
        comment_provider = DisableCommentProvider(s)
        for check in model_serializer_checks:
            yield from check(s, comment_provider)
        comment_provider = DisableMetaCommentProvider(s)
        for check in model_meta_serializer_checks:
            yield from check(s, comment_provider)


class CheckDRFSerializer(BaseCheck):
    @abstractmethod
    def apply(
        self, serializer: Serializer, **kwargs: Any
    ) -> Iterator[django.core.checks.CheckMessage]:
        raise NotImplementedError()


class CheckDRFModelSerializerMeta(BaseCheck):
    @abstractmethod
    def apply(
        self, serializer: ModelSerializer, **kwargs: Any
    ) -> Iterator[django.core.checks.CheckMessage]:
        raise NotImplementedError()


class CheckDRFModelSerializer(BaseCheck):
    @abstractmethod
    def apply(
        self, serializer: ModelSerializer, **kwargs: Any
    ) -> Iterator[django.core.checks.CheckMessage]:
        raise NotImplementedError()


@registry.register("extra_checks_drf_serializer")
class CheckDRFSerializerExtraKwargs(CheckDRFModelSerializerMeta):
    Id = CheckId.X301
    level = django.core.checks.ERROR

    def apply(
        self, serializer: ModelSerializer, **kwargs: Any
    ) -> Iterator[django.core.checks.CheckMessage]:
        if not hasattr(serializer, "Meta") or not hasattr(
            serializer.Meta, "extra_kwargs"
        ):
            return
        invalid = serializer.Meta.extra_kwargs.keys() & serializer._declared_fields
        if invalid:
            yield self.message(
                "extra_kwargs mustn't include fields that declared on serializer.",
                hint=f"Remove extra_kwargs for fields: {', '.join(invalid)}",
                obj=serializer,
            )


@registry.register("extra_checks_drf_serializer")
class CheckDRFSerializerMetaAttribute(CheckDRFModelSerializerMeta):
    Id = CheckId.X302
    settings_form_class = AttrsForm

    def __init__(self, attrs: List[str], **kwargs: Any) -> None:
        self.attrs = attrs
        super().__init__(**kwargs)

    def apply(
        self, serializer: ModelSerializer, **kwargs: Any
    ) -> Iterator[django.core.checks.CheckMessage]:
        meta = getattr(serializer, "Meta", None)
        for attr in self.attrs:
            if not hasattr(meta, attr):
                yield self.message(
                    f"ModelSerializer must define `{attr}` in Meta.",
                    hint=f"Add `{attr}` to serializer's Meta.",
                    obj=serializer,
                )



import inspect
import re
import textwrap
from typing import TYPE_CHECKING, Dict, Iterable, Optional, Set, Type

from extra_checks.check_id import ALL_CHECKS_NAMES, CheckId

if TYPE_CHECKING:
    cached_property = property
else:
    from django.utils.functional import cached_property


DISABLE_COMMENT_PATTERN = r"^#\s*extra-checks-disable-next-line(?:\s+(.*))?$"


def _parse_comment(checks: Optional[str]) -> Set[str]:
    if not checks:
        return ALL_CHECKS_NAMES  # type: ignore
    result = set()
    for scheck in checks.split(","):
        check = CheckId.find_check(scheck.strip())
        if check:
            result.add(check.value)
    return result


def _find_disabled_checks(comments: Iterable[str]) -> Set[str]:
    result = set()
    for line in comments:
        m = re.match(DISABLE_COMMENT_PATTERN, line)
        if m:
            result |= _parse_comment(m.groups()[0])
    return result


class SourceProvider:
    def __init__(self, obj: Type) -> None:
        self._obj = obj
        self._comments_cache: Dict[int, Set[str]] = {}

    @cached_property
    def source(self) -> Optional[str]:
        try:
            return textwrap.dedent(inspect.getsource(self._obj))
        except (TypeError, OSError):
            # TODO: add warning?
            return None

    @cached_property
    def _top_comments(self) -> Iterable[str]:
        for line in (inspect.getcomments(self._obj) or "").splitlines()[::-1]:
            yield line.strip()

    def _get_line_comments(self, line_no: int) -> Iterable[str]:
        line_no -= 2
        if line_no < 0:
            yield from self._top_comments
        for line in (self.source or "").splitlines()[line_no::-1]:
            sline = line.strip()
            if sline.startswith("#"):
                yield sline
            else:
                break

    def get_disabled_checks_for_line(self, line_no: int) -> Set[str]:
        if line_no not in self._comments_cache:
            comments = self._get_line_comments(line_no)
            self._comments_cache[line_no] = _find_disabled_checks(comments)
        return self._comments_cache[line_no]



from typing import Any, Iterable, Iterator, List, Optional

import django.core.checks

from .. import CheckId
from ..registry import ChecksConfig, registry
from .base_checks import BaseCheck


@registry.add_handler("extra_checks_selfcheck")
def check_extra_checks_health(
    checks: Iterable["CheckConfig"],
    config: ChecksConfig,
    app_configs: Optional[List[Any]] = None,
    **kwargs: Any,
) -> Iterator[django.core.checks.CheckMessage]:
    for check in checks:
        yield from check(config, None)


def dict_to_text(data: dict, indent_level: int = 0) -> str:
    output = []
    for field, errors in data.items():
        if not errors:
            continue
        output.append(f"{' '*indent_level * 2}* {field}")
        if isinstance(errors, dict):
            output.append(dict_to_text(errors, indent_level + 1))
        else:
            output.append(
                "\n".join(f"{' '*(indent_level * 2 + 2)}* {e}" for e in errors)
            )
    return "\n".join(output)


@registry.register("extra_checks_selfcheck")
class CheckConfig(BaseCheck):
    Id = CheckId.X001
    level = django.core.checks.CRITICAL

    def apply(
        self, obj: ChecksConfig, **kwargs: Any
    ) -> Iterator[django.core.checks.CheckMessage]:
        if obj.errors:
            yield self.message(
                "Invalid EXTRA_CHECKS config.",
                hint="Fix EXTRA_CHECKS in your settings. Errors:\n"
                + (dict_to_text(obj.errors) if obj.errors else "No details."),
            )



import enum
from typing import Optional, cast


class CheckId(str, enum.Enum):
    X001 = "extra-checks-config"
    X010 = "model-attribute"
    X011 = "model-meta-attribute"
    X012 = "model-admin"
    X013 = "no-unique-together"
    X014 = "no-index-together"
    X050 = "field-verbose-name"
    X051 = "field-verbose-name-gettext"
    X052 = "field-verbose-name-gettext-case"
    X053 = "field-help-text-gettext"
    X054 = "field-file-upload-to"
    X055 = "field-text-null"
    X056 = "field-boolean-null"
    X057 = "field-null"
    X058 = "field-foreign-key-db-index"
    X059 = "field-default-null"
    X060 = "field-choices-constraint"
    X061 = "field-related-name"
    X301 = "drf-model-serializer-extra-kwargs"
    X302 = "drf-model-serializer-meta-attribute"

    @classmethod
    def find_check(cls, value: str) -> Optional["CheckId"]:
        try:
            return cls(value)
        except ValueError:
            pass
        try:
            return cast(CheckId, cls._member_map_[value])
        except KeyError:
            pass
        return None


ALL_CHECKS_NAMES = frozenset(CheckId._value2member_map_.keys())



from extra_checks import CheckId

SECRET_KEY = "random"

ALLOWED_HOSTS = ["*"]
USE_TZ = True

INSTALLED_APPS = [
    "django.contrib.sites",
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.messages",
    "django.contrib.sessions",
    "tests.example",
    "extra_checks",
]

EXTRA_CHECKS = {
    "include_apps": ["tests.example"],
    "level": "ERROR",
    "checks": [
        # require non empty `upload_to` argument.
        "field-file-upload-to",
        # use dict form if check need configuration
        # eg. all models must have fk to Site model
        {"id": "model-attribute", "attrs": ["site"]},
        # require `db_table` for all models, increase level to ERROR
        {"id": "model-meta-attribute", "attrs": ["db_table"], "level": "ERROR"},
        {"id": "drf-model-serializer-meta-attribute", "attrs": ["read_only_fields"]},
    ],
}

_checks = [c["id"] if isinstance(c, dict) else c for c in EXTRA_CHECKS["checks"]]
EXTRA_CHECKS["checks"].extend(list(CheckId._value2member_map_.keys() - _checks))  # type: ignore

DATABASES = {"default": {"ENGINE": "django.db.backends.sqlite3"}}
TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "OPTIONS": {
            "context_processors": [
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    }
]

MIDDLEWARE = [
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
]

ROOT_URLCONF = "tests.urls"

SILENCED_SYSTEM_CHECKS = ["fields.E210"]



import pytest

from extra_checks.checks import model_checks, model_field_checks
from tests.example import models


@pytest.fixture
def test_case(test_case):
    return test_case.handler(model_checks.check_models)


def test_check_field_verbose_name(test_case):
    messages = (
        test_case.settings(
            {"checks": [model_field_checks.CheckFieldVerboseName.Id.value]}
        )
        .models(models.ModelFieldVerboseName)
        .check(model_field_checks.CheckFieldVerboseName)
        .run()
    )
    assert {m.obj.name for m in messages} == {
        "no_name",
        "no_name_related",
        "no_name_nested_field",
    }


def test_check_field_verbose_name_gettext(test_case):
    messages = (
        test_case.models(models.ModelFieldVerboseName)
        .settings(
            {"checks": [model_field_checks.CheckFieldVerboseNameGettext.Id.value]}
        )
        .check(model_field_checks.CheckFieldVerboseNameGettext)
        .run()
    )
    assert {m.obj.name for m in messages} == {
        "first_arg_name",
        "kwarg_name",
        "gettext",
        "name_related",
        "nested_field",
    }


def test_check_field_verbose_name_gettext_case(test_case):
    messages = (
        test_case.models(models.ModelFieldVerboseName)
        .settings(
            {"checks": [model_field_checks.CheckFieldVerboseNameGettextCase.Id.value]}
        )
        .check(model_field_checks.CheckFieldVerboseNameGettextCase)
        .run()
    )
    assert len(messages) == 1
    assert messages[0].id == model_field_checks.CheckFieldVerboseNameGettextCase.Id.name
    assert {m.obj.name for m in messages} == {
        "gettext_case",
    }


def test_check_field_null_boolean(test_case):
    messages = (
        test_case.models(models.ModelFieldNullFalse)
        .settings({"checks": [model_field_checks.CheckFieldNullBoolean.Id.value]})
        .check(model_field_checks.CheckFieldNullBoolean)
        .run()
    )
    assert len(messages) == 1
    assert messages[0].id == model_field_checks.CheckFieldNullBoolean.Id.name
    assert {m.obj.name for m in messages} == {
        "null_fail",
    }


def test_check_field_help_text_gettext(test_case):
    messages = (
        test_case.settings(
            {"checks": [model_field_checks.CheckFieldHelpTextGettext.Id.value]}
        )
        .models(models.ModelFieldHelpTextGettext)
        .check(model_field_checks.CheckFieldHelpTextGettext)
        .run()
    )
    assert len(messages) == 1
    assert messages[0].id == model_field_checks.CheckFieldHelpTextGettext.Id.name
    assert {m.obj.name for m in messages} == {
        "text_fail",
    }


def test_check_field_file_upload_to(test_case):
    messages = (
        test_case.models(models.ModelFieldFileUploadTo)
        .settings({"checks": [model_field_checks.CheckFieldFileUploadTo.Id.value]})
        .check(model_field_checks.CheckFieldFileUploadTo)
        .run()
    )
    assert {m.obj.name for m in messages} == {
        "image_fail",
        "file_fail",
    }


def test_check_field_text_null(test_case):
    messages = (
        test_case.settings({"checks": [model_field_checks.CheckFieldTextNull.Id.value]})
        .models(models.ModelFieldTextNull)
        .check(model_field_checks.CheckFieldTextNull)
        .run()
    )
    assert {m.obj.name for m in messages} == {
        "text_fail",
        "chars_fail",
        "custom_fail",
    }


def test_check_field_null_false(test_case):
    messages = (
        test_case.settings(
            {"checks": [model_field_checks.CheckFieldNullFalse.Id.value]}
        )
        .models(models.ModelFieldNullFalse)
        .check(model_field_checks.CheckFieldNullFalse)
        .run()
    )
    assert {m.obj.name for m in messages} == {
        "myfield_fail",
    }


def test_check_field_foreign_key_index(test_case):
    messages = (
        test_case.settings(
            {"checks": [model_field_checks.CheckFieldForeignKeyIndex.Id.value]}
        )
        .models(models.ModelFieldForeignKeyIndex)
        .check(model_field_checks.CheckFieldForeignKeyIndex)
        .run()
    )
    assert {m.obj.name for m in messages} == {
        "field_one",
        "author",
        "field_three",
        "field_in_indexes",
        "field_index_desc",
    }


def test_check_field_foreign_key_index_always(test_case):
    settings = {
        "checks": [
            {
                "id": model_field_checks.CheckFieldForeignKeyIndex.Id.value,
                "when": "always",
            }
        ]
    }
    messages = (
        test_case.settings(settings)
        .models(models.ModelFieldForeignKeyIndex)
        .check(model_field_checks.CheckFieldForeignKeyIndex)
        .run()
    )
    assert {m.obj.name for m in messages} == {
        "field_one",
        "author",
        "another_article",
        "field_three",
        "field_in_indexes",
        "field_index_desc",
    }


def test_check_field_related_name(test_case):
    messages = (
        test_case.settings(
            {"checks": [model_field_checks.CheckFieldRelatedName.Id.value]}
        )
        .models(models.Article)
        .check(model_field_checks.CheckFieldRelatedName)
        .run()
    )
    assert {m.obj.name for m in messages} == {
        "site",
    }


def test_generic_key_null(test_case):
    """ensure custom fields without null attribute not checked by rule"""
    messages = (
        test_case.settings(
            {"checks": [model_field_checks.CheckFieldNullFalse.Id.value]}
        )
        .models(models.GenericKeyOne, models.GenericKeyTwo)
        .check(model_field_checks.CheckFieldNullFalse)
        .run()
    )
    assert not messages


def test_verbose_name_of_related_field(test_case):
    """if field inherit from RelatedField the first argument is not a verbose name."""
    messages = (
        test_case.settings(
            {"checks": [model_field_checks.CheckFieldVerboseNameGettext.Id.value]}
        )
        .models(models.GenericKeyTwo)
        .check(model_field_checks.CheckFieldVerboseNameGettext)
        .run()
    )
    assert not messages


def test_field_null_default_null(test_case):
    messages = (
        test_case.settings(
            {"checks": [model_field_checks.CheckFieldDefaultNull.Id.value]}
        )
        .models(models.ModelFieldNullDefault)
        .check(model_field_checks.CheckFieldDefaultNull)
        .run()
    )
    assert len(messages) == 1
    assert messages[0].obj.name == "myfield_fail"


def test_field_choices_constraint(test_case):
    messages = (
        test_case.settings(
            {"checks": [model_field_checks.CheckFieldChoicesConstraint.Id.value]}
        )
        .models(models.ChoicesConstraint)
        .check(model_field_checks.CheckFieldChoicesConstraint)
        .run()
    )
    errors = {m.obj.name: m for m in messages}
    assert errors.keys() == {
        "partial",
        "missed",
        "blank_missed",
        "blank_included",
        "integer_blank_invalid",
    }
    assert 'check=models.Q(partial__in=["S", "C"]))' in errors["partial"].hint
    assert "check=models.Q(missed__in=[1, 2]))" in errors["missed"].hint
    assert (
        'check=models.Q(blank_missed__in=["A", "B", ""])' in errors["blank_missed"].hint
    )
    assert (
        'check=models.Q(blank_included__in=["A", "B", ""])'
        in errors["blank_included"].hint
    )
    assert (
        "check=models.Q(integer_blank_invalid__in=[1, 2])"
        in errors["integer_blank_invalid"].hint
    )



import typing

import django.core.checks
from django import forms
from django.utils.translation import gettext_lazy as _

from . import CheckId


class ListField(forms.Field):
    default_error_messages = {
        "invalid_list": _("Enter a list of values."),
    }

    def __init__(self, base_field: forms.Field, **kwargs: typing.Any) -> None:
        self.base_field = base_field
        super().__init__(**kwargs)

    def to_python(self, value: typing.Any) -> list:
        if not value:
            return []
        if not isinstance(value, (list, tuple)):
            raise forms.ValidationError(
                self.error_messages["invalid_list"], code="invalid_list"
            )
        return [self.base_field.to_python(val) for val in value]

    def validate(self, value: list) -> None:
        if self.required and not value:
            raise forms.ValidationError(
                self.error_messages["required"], code="required"
            )
        for val in value:
            self.base_field.validate(val)


class FilterField(forms.Field):
    default_error_messages = {
        "invalid_callable": _("%(value)s is not valid callable for skipif."),
    }

    def to_python(self, value: typing.Any) -> typing.Optional[typing.Callable]:
        if not value:
            return None
        if not callable(value):
            raise forms.ValidationError(
                self.error_messages["invalid_callable"], code="invalid_callable"
            )
        return value


class UnionField(forms.Field):
    default_error_messages = {
        "type_invalid": _("%(value)s is not one of the available types."),
    }

    def __init__(
        self, base_fields: typing.Dict[typing.Any, forms.Field], **kwargs: typing.Any
    ) -> None:
        assert isinstance(base_fields, dict)
        self.base_fields = base_fields
        super().__init__(**kwargs)

    def to_python(self, value: typing.Any) -> typing.Any:
        for type_, field in self.base_fields.items():
            if isinstance(value, type_):
                return field.to_python(value)
        raise forms.ValidationError(
            self.error_messages["type_invalid"],
            code="type_invalid",
            params={"value": value},
        )

    def validate(self, value: typing.Any) -> None:
        for type_, field in self.base_fields.items():
            if isinstance(value, type_):
                field.validate(value)
                return


class DictField(forms.ChoiceField):
    default_error_messages = {
        "invalid_choice": _("ID %(value)s is not one of the available checks."),
        "invalid_dict": _("Must be a dict."),
        "id_required": _("`id` field is required."),
    }

    def __init__(
        self, id_choices: typing.List[typing.Tuple[str, str]], **kwargs: typing.Any
    ) -> None:
        super().__init__(choices=id_choices, **kwargs)

    def to_python(self, value: typing.Any) -> dict:
        if not value:
            return {}
        if not isinstance(value, dict):
            raise forms.ValidationError(
                self.error_messages["invalid_dict"],
                code="invalid_dict",
            )
        return {str(k): v for k, v in value.items()}

    def validate(self, value: dict) -> None:
        if self.required and not value:
            raise forms.ValidationError(
                self.error_messages["required"], code="required"
            )
        if "id" not in value:
            raise forms.ValidationError(
                self.error_messages["id_required"], code="id_required"
            )
        if not self.valid_value(value["id"]):
            raise forms.ValidationError(
                self.error_messages["invalid_choice"],
                code="invalid_choice",
                params={"value": value["id"]},
            )


class ConfigForm(forms.Form):
    include_apps = ListField(forms.CharField(), required=False)
    level = forms.ChoiceField(
        choices=[(c, c) for c in ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]],
        required=False,
    )
    checks = ListField(
        UnionField(
            {
                str: forms.ChoiceField(
                    choices=[(v.value, v.value) for v in CheckId.__members__.values()],
                    error_messages={
                        "invalid_choice": _(
                            "%(value)s is not one of the available checks."
                        ),
                    },
                ),
                dict: DictField(
                    id_choices=[
                        (v.value, v.value) for v in CheckId.__members__.values()
                    ]
                ),
            }
        ),
        required=False,
    )

    def clean_checks(self) -> typing.Dict[str, dict]:
        result: typing.Dict[str, dict] = {}
        for check in self.cleaned_data["checks"]:
            if isinstance(check, str):
                result[check] = {}
            else:
                result[check["id"]] = check
        return result

    def clean(self) -> typing.Dict[str, typing.Any]:
        if (
            "include_apps" in self.cleaned_data
            and not self.cleaned_data["include_apps"]
            and "include_apps" not in self.data
        ):
            del self.cleaned_data["include_apps"]
        if "level" in self.cleaned_data and "checks" in self.cleaned_data:
            for check in self.cleaned_data["checks"].values():
                check.setdefault("level", self.cleaned_data["level"])
            del self.cleaned_data["level"]
        return self.cleaned_data

    def is_valid(self, check_forms: typing.Dict[CheckId, "typing.Type[BaseCheckForm]"]) -> bool:  # type: ignore
        if not super().is_valid():
            return False
        checks = self.cleaned_data.get("checks", {})
        rforms = {
            name: check_forms[name](data=value)
            for name, value in checks.items()
            if name in check_forms
        }
        if forms.all_valid(rforms.values()):  # type: ignore
            self.cleaned_data["checks"] = {
                name: f.cleaned_data for name, f in rforms.items()
            }
            return True
        self.errors["checks"] = {name: form.errors for name, form in rforms.items()}
        return False


class BaseCheckForm(forms.Form):
    level = forms.ChoiceField(
        choices=[(c, c) for c in ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]],
        required=False,
    )
    skipif = FilterField(required=False)

    def clean_level(self) -> typing.Optional[int]:
        if self.cleaned_data["level"]:
            return getattr(django.core.checks, self.cleaned_data["level"])
        return None

    def clean(self) -> typing.Dict[str, typing.Any]:
        if "skipif" in self.cleaned_data and not self.cleaned_data["skipif"]:
            del self.cleaned_data["skipif"]
        return self.cleaned_data


class AttrsForm(BaseCheckForm):
    attrs = ListField(forms.CharField())



from abc import abstractmethod
from typing import Any, Iterator, Optional, Type

import django.core.checks
from django import forms
from django.db import models

from .. import CheckId
from ..ast import FieldASTProtocol, MissingASTError
from ..ast.protocols import DisableCommentProtocol
from ..forms import BaseCheckForm
from ..registry import registry
from .base_checks import BaseCheck, BaseCheckMixin


class CheckModelField(BaseCheck):
    @abstractmethod
    def apply(
        self,
        field: models.fields.Field,
        *,
        ast: FieldASTProtocol,
        model: Type[models.Model],
    ) -> Iterator[django.core.checks.CheckMessage]:
        raise NotImplementedError()

    def __call__(
        self, obj: Any, ast: Optional[DisableCommentProtocol] = None, **kwargs: Any
    ) -> Iterator[django.core.checks.CheckMessage]:
        try:
            yield from super().__call__(obj, ast=ast, **kwargs)
        except MissingASTError:
            pass

    def is_ignored(self, obj: Any) -> bool:
        if self.skipif and self.skipif(obj):
            return True
        return obj.model in self.ignore_objects or type(obj) in self.ignore_types


class GetTextMixin(BaseCheckMixin):
    class GettTextFuncForm(BaseCheckForm):
        gettext_func = forms.CharField(required=False)

    settings_form_class = GettTextFuncForm

    def __init__(self, gettext_func: str, **kwargs: Any) -> None:
        self.gettext_func = gettext_func or "_"
        super().__init__(**kwargs)


@registry.register(django.core.checks.Tags.models)
class CheckFieldVerboseName(CheckModelField):
    Id = CheckId.X050

    def apply(
        self, field: models.fields.Field, ast: FieldASTProtocol, **kwargs: Any
    ) -> Iterator[django.core.checks.CheckMessage]:
        if not ast.get_arg("verbose_name"):
            yield self.message(
                "Field has no verbose name.",
                hint="Set verbose name on the field.",
                obj=field,
            )


@registry.register(django.core.checks.Tags.models)
class CheckFieldVerboseNameGettext(GetTextMixin, CheckModelField):
    Id = CheckId.X051

    def apply(
        self, field: models.fields.Field, ast: FieldASTProtocol, **kwargs: Any
    ) -> Iterator[django.core.checks.CheckMessage]:
        verbose_name = ast.get_arg("verbose_name")
        if verbose_name and not (
            verbose_name.is_callable
            and verbose_name.callable_func_name == self.gettext_func
        ):
            yield self.message(
                "Verbose name should use gettext.",
                hint="Use gettext on the verbose name.",
                obj=field,
            )


@registry.register(django.core.checks.Tags.models)
class CheckFieldVerboseNameGettextCase(GetTextMixin, CheckModelField):
    Id = CheckId.X052

    def apply(
        self, field: models.fields.Field, ast: FieldASTProtocol, **kwargs: Any
    ) -> Iterator[django.core.checks.CheckMessage]:
        verbose_name = ast.get_arg("verbose_name")
        if verbose_name and (
            verbose_name.is_callable
            and verbose_name.callable_func_name == self.gettext_func
        ):
            value = verbose_name.get_call_first_args()
            if (
                value
                and isinstance(value, str)
                and not all(
                    w.islower() or w.isupper() or w.isdigit() for w in value.split(" ")
                )
            ):
                yield self.message(
                    "Words in verbose name must be all upper case or all lower case.",
                    hint=f'Change verbose name to "{value.lower()}".',
                    obj=field,
                )


@registry.register(django.core.checks.Tags.models)
class CheckFieldHelpTextGettext(GetTextMixin, CheckModelField):
    Id = CheckId.X053

    def apply(
        self, field: models.fields.Field, ast: FieldASTProtocol, **kwargs: Any
    ) -> Iterator[django.core.checks.CheckMessage]:
        help_text = ast.get_arg("help_text")
        if help_text and not (
            help_text.is_callable and help_text.callable_func_name == self.gettext_func
        ):
            yield self.message(
                "Help text should use gettext.",
                hint="Use gettext on the help text.",
                obj=field,
            )


@registry.register(django.core.checks.Tags.models)
class CheckFieldFileUploadTo(CheckModelField):
    Id = CheckId.X054

    def apply(
        self, field: models.fields.Field, **kwargs: Any
    ) -> Iterator[django.core.checks.CheckMessage]:
        if isinstance(field, models.FileField):
            if not field.upload_to:
                yield self.message(
                    f'Field "{field.name}" must have non empty "upload_to" attribute.',
                    hint='Set "upload_to" on the field.',
                    obj=field,
                )


@registry.register(django.core.checks.Tags.models)
class CheckFieldTextNull(CheckModelField):
    Id = CheckId.X055

    def apply(
        self, field: models.fields.Field, **kwargs: Any
    ) -> Iterator[django.core.checks.CheckMessage]:
        if isinstance(field, (models.CharField, models.TextField)):
            if field.null:
                yield self.message(
                    f'Field "{field.name}" shouldn\'t use `null=True` '
                    "(django uses empty string for text fields).",
                    hint="Remove `null=True` attribute from the field.",
                    obj=field,
                )


@registry.register(django.core.checks.Tags.models)
class CheckFieldNullBoolean(CheckModelField):
    Id = CheckId.X056
    deprecation_warnings = [
        "`field-boolean-null` check is deprecated and will be removed in version 0.14.0"
    ]

    def apply(
        self, field: models.fields.Field, **kwargs: Any
    ) -> Iterator[django.core.checks.CheckMessage]:
        if isinstance(field, models.NullBooleanField):
            yield self.message(
                f'Field "{field.name}" should be `BooleanField` with attribute `null=True`.',
                hint="Replace `NullBooleanField` by `BooleanField` with attribute `null=True`.",
                obj=field,
            )


@registry.register(django.core.checks.Tags.models)
class CheckFieldNullFalse(CheckModelField):
    Id = CheckId.X057

    def apply(
        self, field: models.fields.Field, ast: FieldASTProtocol, **kwargs: Any
    ) -> Iterator[django.core.checks.CheckMessage]:
        if field.null is False and ast.get_arg("null"):
            yield self.message(
                "Argument `null=False` is default.",
                hint="Remove `null=False` from field arguments.",
                obj=field,
            )


@registry.register(django.core.checks.Tags.models)
class CheckFieldForeignKeyIndex(CheckModelField):
    Id = CheckId.X058

    class CheckFieldForeignKeyIndexForm(BaseCheckForm):
        when = forms.ChoiceField(
            choices=[("indexes", "indexes"), ("always", "always")],
            required=False,
        )

    settings_form_class = CheckFieldForeignKeyIndexForm

    def __init__(self, when: str, **kwargs: Any) -> None:
        self.when = when or "indexes"
        super().__init__(**kwargs)

    @classmethod
    def get_index_values_in_meta(cls, model: Type[models.Model]) -> Iterator[str]:
        for entry in model._meta.unique_together:
            yield from entry
        for entry in model._meta.index_together:
            yield from entry
        for constraint in model._meta.constraints:
            if isinstance(constraint, models.UniqueConstraint):
                yield from constraint.fields
        for index in model._meta.indexes:
            yield from index.fields

    @classmethod
    def get_fields_with_indexes_in_meta(
        cls, model: Type[models.Model]
    ) -> Iterator[str]:
        for entry in cls.get_index_values_in_meta(model):
            yield entry.lstrip("-")

    def apply(
        self,
        field: models.fields.Field,
        ast: FieldASTProtocol,
        model: Type[models.Model],
    ) -> Iterator[django.core.checks.CheckMessage]:
        if isinstance(field, models.fields.related.RelatedField):
            if field.many_to_one and not ast.get_arg("db_index"):
                if self.when == "indexes":
                    if field.name in self.get_fields_with_indexes_in_meta(model):
                        yield self.message(
                            "ForeignKey field must set `db_index` explicitly if it present in other indexes.",
                            hint="Specify `db_index` field argument.",
                            obj=field,
                        )
                else:
                    yield self.message(
                        "ForeignKey must set `db_index` explicitly.",
                        hint="Specify `db_index` field argument.",
                        obj=field,
                    )


@registry.register(django.core.checks.Tags.models)
class CheckFieldRelatedName(CheckModelField):
    Id = CheckId.X061

    def apply(
        self,
        field: models.fields.Field,
        ast: FieldASTProtocol,
        model: Type[models.Model],
    ) -> Iterator[django.core.checks.CheckMessage]:
        if isinstance(field, models.fields.related.RelatedField):
            if not field.remote_field.related_name:
                yield self.message(
                    "Related fields must set `related_name` explicitly.",
                    hint="Specify `related_name` field argument. Use `related_name='+'` to not create a backwards relation.",
                    obj=field,
                )


@registry.register(django.core.checks.Tags.models)
class CheckFieldDefaultNull(CheckModelField):
    Id = CheckId.X059

    def apply(
        self, field: models.fields.Field, ast: FieldASTProtocol, **kwargs: Any
    ) -> Iterator[django.core.checks.CheckMessage]:
        if field.null and field.default is None and ast.get_arg("default"):
            yield self.message(
                "Argument `default=None` is redundant if `null=True` is set. (see docs about exceptions).",
                hint="Remove `default=None` from field arguments.",
                obj=field,
            )


@registry.register(django.core.checks.Tags.models)
class CheckFieldChoicesConstraint(CheckModelField):
    Id = CheckId.X060

    @staticmethod
    def _repr_choice(value: Any) -> str:
        if isinstance(value, str):
            return f'"{value}"'
        return str(value)

    def apply(
        self,
        field: models.fields.Field,
        ast: FieldASTProtocol,
        model: Type[models.Model],
    ) -> Iterator[django.core.checks.CheckMessage]:
        choices = field.flatchoices
        if choices:
            field_choices = [c[0] for c in choices]
            if field.empty_strings_allowed and field.blank and "" not in field_choices:
                field_choices.append("")
            in_name = f"{field.name}__in"
            for constraint in model._meta.constraints:
                if isinstance(constraint, models.CheckConstraint):
                    conditions = dict(constraint.check.children)  # type: ignore
                    if in_name in conditions and set(field_choices) == set(
                        conditions[in_name]
                    ):
                        return
            check = f'models.Q({in_name}=[{", ".join([self._repr_choice(c) for c in field_choices])}])'
            yield self.message(
                "Field with choices must have companion CheckConstraint to enforce choices on database level.",
                hint=f'Add to Meta.constraints: `models.CheckConstraint(name="%(app_label)s_%(class)s_{field.name}_valid", check={check})`',
                obj=field,
            )



from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
)

import django.apps
import django.core.checks
from django import forms
from django.conf import settings

from . import CheckId
from .forms import ConfigForm

if TYPE_CHECKING:
    from .checks import BaseCheck


class ChecksConfig:
    def __init__(
        self,
        *,
        errors: Optional[forms.utils.ErrorDict] = None,
        checks: Optional[Dict[CheckId, dict]] = None,
        include_apps: Optional[Iterable[str]] = None,
        ignored_objects: Optional[Dict[CheckId, Set[Any]]] = None,
    ) -> None:
        self.checks: Dict[CheckId, dict] = {**(checks or {}), CheckId.X001: {}}
        self.include_apps = include_apps
        self.errors = errors
        self.ignored_objects: Dict[CheckId, set] = ignored_objects or {}

    @classmethod
    def create(
        cls,
        include_checks: Dict["Type[BaseCheck]", Sequence[str]],
        ignore_checks: Optional[Dict[Any, Set[Union[str, CheckId]]]] = None,
    ) -> "ChecksConfig":
        check_forms = {r.Id: r.settings_form_class for r in include_checks}
        if not hasattr(settings, "EXTRA_CHECKS"):
            return cls()
        form = ConfigForm(settings.EXTRA_CHECKS)
        if not form.is_valid(check_forms):
            return cls(errors=form.errors)
        ignored, errors = ChecksConfig._build_ignored(ignore_checks or {})
        if errors:
            return cls(errors={"__all__": errors})  # type: ignore
        return cls(ignored_objects=ignored, **form.cleaned_data)

    @staticmethod
    def _build_ignored(
        ignore_checks: Dict[Any, Set[Union[str, CheckId]]]
    ) -> Tuple[Dict[CheckId, Set[Any]], List[str]]:
        errors = []
        ignored: Dict[CheckId, set] = {}
        for obj, ids in ignore_checks.items():
            for id_ in ids:
                check_id = CheckId.find_check(id_)
                if check_id:
                    ignored.setdefault(check_id, set()).add(obj)
                else:
                    errors.append(
                        f"Unknown check ({id_}) provided to the 'ignore_checks'."
                    )
        return ignored, errors


_ChecksHandler = Callable[[Optional[List[Any]], Any], Iterator[Any]]


class Registry:
    def __init__(self) -> None:
        self.registered_checks: Dict["Type[BaseCheck]", Sequence[str]] = {}
        self.enabled_checks: Dict[str, List["BaseCheck"]] = {}
        self.ignored_checks: Dict[Any, Set[Union[CheckId, str]]] = {}
        self.handlers: Dict[str, _ChecksHandler] = {}
        self._config: Optional[ChecksConfig] = None

    def _register(
        self, tags: List[str], check_class: "Type[BaseCheck]"
    ) -> "Type[BaseCheck]":
        self.registered_checks[check_class] = tags
        return check_class

    def _add_handler(self, tag: str, handler: _ChecksHandler) -> _ChecksHandler:
        self.handlers[tag] = handler
        return handler

    def _bind_handler(
        self,
        tag: str,
        handler: _ChecksHandler,
        checks: List["BaseCheck"],
        config: ChecksConfig,
    ) -> Optional[Callable]:
        if checks:
            f = partial(handler, checks, config)
            django.core.checks.register(f, tag)
            return f
        return None

    def register(self, *tags: str) -> Callable[["Type[BaseCheck]"], "Type[BaseCheck]"]:
        return partial(self._register, tags)

    def add_handler(self, tag: str) -> Callable[[Callable], Callable]:
        return partial(self._add_handler, tag)

    def bind(self) -> Dict[str, Callable]:
        config = ChecksConfig.create(self.registered_checks, self.ignored_checks)
        for check_class, tags in self.registered_checks.items():
            if check_class.Id in config.checks:
                check = check_class(
                    ignore_objects=config.ignored_objects.get(check_class.Id, set()),
                    **config.checks[check_class.Id],
                )
                for tag in tags:
                    self.enabled_checks.setdefault(tag, []).append(check)
        tag_handlers = {}
        for tag, handler in self.handlers.items():
            bh = self._bind_handler(
                tag, handler, self.enabled_checks.get(tag, []), config
            )
            if bh:
                tag_handlers[tag] = bh
        self._config = config
        return tag_handlers

    @property
    def is_healthy(self) -> bool:
        return True if self._config is None else not self._config.errors


registry = Registry()



import textwrap

import django.core.checks
import pytest

from extra_checks.check_id import CheckId
from extra_checks.checks import (
    CheckConfig,
    CheckFieldFileUploadTo,
    CheckFieldForeignKeyIndex,
    check_extra_checks_health,
)
from extra_checks.checks.base_checks import BaseCheck
from extra_checks.registry import Registry
from extra_checks.utils import collect_subclasses


def test_empty_config(registry, settings):
    settings.EXTRA_CHECKS = {}
    registry.bind()
    assert registry.is_healthy
    assert len(registry.registered_checks) == 1
    assert len(registry.enabled_checks) == 1

    settings.EXTRA_CHECKS = {"checks": []}
    registry = Registry()
    registry._register(["extra_checks_selfcheck"], CheckConfig)
    registry.bind()
    assert registry.is_healthy
    assert len(registry.registered_checks) == 1
    assert len(registry.enabled_checks) == 1


def test_error_formatting(registry, settings):
    settings.EXTRA_CHECKS = {
        "checks": [
            CheckFieldFileUploadTo.Id.value,
            {"id": CheckFieldForeignKeyIndex.Id.value, "when": "random"},
        ]
    }
    registry._register([django.core.checks.Tags.models], CheckFieldForeignKeyIndex)
    registry._register([django.core.checks.Tags.models], CheckFieldFileUploadTo)
    registry._add_handler("extra_checks_selfcheck", check_extra_checks_health)
    handlers = registry.bind()
    assert not registry.is_healthy
    messages = list(handlers["extra_checks_selfcheck"]())
    assert len(messages) == 1
    assert (
        messages[0].hint
        == textwrap.dedent(
            f"""
            Fix EXTRA_CHECKS in your settings. Errors:
            * checks
              * {CheckFieldForeignKeyIndex.Id.value}
                * when
                  * Select a valid choice. random is not one of the available choices.
            """
        ).strip()
    )


def test_unique_check_ids():
    pytest.importorskip("rest_framework")
    used_checks = [
        c.Id for c in collect_subclasses(BaseCheck.__subclasses__()) if hasattr(c, "Id")
    ]
    unused = CheckId._value2member_map_.keys() - set(used_checks)
    assert not unused, "Not all CheckIds used."
    dups = {c.value for c in used_checks if used_checks.count(c) > 1}
    assert not dups, "CheckIds must be unique per Check."



import pytest

from extra_checks.checks import model_checks
from tests.example.models import (
    Article,
    Author,
    ModelFieldForeignKeyIndex,
    ModelFieldTextNull,
)


@pytest.fixture
def test_case(test_case):
    return test_case.handler(model_checks.check_models)


def test_get_models_to_check():
    models = list(model_checks._get_models_to_check())
    assert models
    for model in models:
        assert model._meta.app_label == "example"


def test_get_models_to_check_include_apps():
    models = list(model_checks._get_models_to_check(include_apps=[]))
    assert not models

    models = list(
        model_checks._get_models_to_check(include_apps=["django.contrib.sites"])
    )
    assert models
    for model in models:
        assert model._meta.app_label == "sites"


def test_check_model_attrs(test_case):
    messages = (
        test_case.models(Article)
        .settings(
            {
                "checks": [
                    {"id": model_checks.CheckModelAttribute.Id.value, "attrs": ["site"]}
                ]
            }
        )
        .check(model_checks.CheckModelAttribute)
        .run()
    )
    assert not messages
    messages = test_case.models(Author).run()
    assert len(messages) == 1
    assert messages[0].id == model_checks.CheckModelAttribute.Id.name


def test_check_model_meta_attrs(test_case):
    messages = (
        test_case.models(Article)
        .settings(
            {
                "checks": [
                    {
                        "id": model_checks.CheckModelMetaAttribute.Id.value,
                        "attrs": ["verbose_name"],
                    }
                ]
            }
        )
        .check(model_checks.CheckModelMetaAttribute)
        .run()
    )
    assert not messages
    messages = test_case.models(Author).run()
    assert len(messages) == 1
    assert messages[0].id == model_checks.CheckModelMetaAttribute.Id.name


def test_admin_models(test_case):
    messages = (
        test_case.models(Article, Author)
        .settings({"checks": [model_checks.CheckModelAdmin.Id.value]})
        .check(model_checks.CheckModelAdmin)
        .run()
    )
    assert not messages
    messages = test_case.models(ModelFieldTextNull).run()
    assert len(messages) == 1
    assert messages[0].id == model_checks.CheckModelAdmin.Id.name


def test_no_unique_together(test_case):
    messages = (
        test_case.models(ModelFieldForeignKeyIndex)
        .settings({"checks": [model_checks.CheckNoUniqueTogether.Id.value]})
        .check(model_checks.CheckNoUniqueTogether)
        .run()
    )
    assert len(messages) == 1
    assert messages[0].id == model_checks.CheckNoUniqueTogether.Id.name


def test_no_index_together(test_case):
    messages = (
        test_case.models(ModelFieldForeignKeyIndex)
        .settings({"checks": [model_checks.CheckNoIndexTogether.Id.value]})
        .check(model_checks.CheckNoIndexTogether)
        .run()
    )
    assert len(messages) == 1
    assert messages[0].id == model_checks.CheckNoIndexTogether.Id.name
